from starlette.routing import Route
from starlette.routing import WebSocketRoute

from .views import WsChatEndpoint
from .views import Chat
from .views import Chat1
from .views import Chat2
from .views import SendMessageFromAnyPartOfYourCode
from .views import ShowGroups
from .views import FlushGroups
from .views import ShowHistory
from .views import FlushHistory
from .views import CleanExpired


routes = [
    WebSocketRoute("/chat_ws", WsChatEndpoint),
    Route("/", endpoint=Chat),
    Route("/chat1", endpoint=Chat1),
    Route("/chat2", endpoint=Chat2),
    Route(
        "/send-message-from-any-part-of-your-code",
        endpoint=SendMessageFromAnyPartOfYourCode,
    ),
    Route("/show-groups", endpoint=ShowGroups),
    Route("/flush-groups", endpoint=FlushGroups),
    Route("/show-history", endpoint=ShowHistory),
    Route("/flush-history", endpoint=FlushHistory),
    Route("/clean-expired", endpoint=CleanExpired),
]



import pytest

from unittest.mock import MagicMock
from channel_box import Channel, ChannelBox
from starlette.websockets import WebSocket


@pytest.fixture
def mock_websocket():
    mock = MagicMock(spec=WebSocket)
    mock.send.return_value = None
    return mock


@pytest.mark.asyncio
async def test_channel_box(
    mock_websocket,
):
    group_name = "test_group"

    channel = Channel(
        websocket=mock_websocket,
        expires=60 * 60,
        payload_type="json",
    )

    await ChannelBox.add_channel_to_group(
        channel=channel,
        group_name=group_name,
    )
    groups = await ChannelBox.get_groups()
    assert group_name in groups
    assert "created_at" in groups[group_name][channel]

    await ChannelBox.remove_channel_from_group(
        channel=channel,
        group_name=group_name,
    )
    await ChannelBox.add_channel_to_group(
        channel=channel,
        group_name=group_name,
    )
    groups = await ChannelBox.get_groups()
    assert groups
    assert len(groups) == 1

    await ChannelBox.flush_groups()
    groups = await ChannelBox.get_groups()
    assert not groups

    await ChannelBox.add_channel_to_group(
        channel=channel,
        group_name=group_name,
    )

    await ChannelBox.group_send(
        group_name=group_name,
        payload={"hello": "world"},
        save_history=True,
    )

    history = await ChannelBox.get_history()
    assert history
    assert len(history) == 1

    await ChannelBox.flush_history()

    history = await ChannelBox.get_history()
    assert not history



from datetime import UTC, datetime
from enum import Enum
from dataclasses import dataclass
import uuid
from uuid import UUID


class PayloadTypeEnum(Enum):
    JSON = "json"
    TEXT = "text"
    BYTES = "bytes"


@dataclass
class ChannelMessageDC:
    payload: str | bytes | dict
    uuid: UUID = uuid.uuid4()
    created: datetime = datetime.now(tz=UTC)



import json

from simple_print import sprint
from jinja2 import Template

from channel_box import Channel
from channel_box import ChannelBox

from .utils import main_template
from .utils import chat_template

from starlette.endpoints import WebSocketEndpoint
from starlette.responses import JSONResponse
from starlette.endpoints import HTTPEndpoint
from starlette.responses import HTMLResponse


class WsChatEndpoint(WebSocketEndpoint):
    async def on_connect(
        self,
        websocket,
    ):
        sprint(
            f"{self.__class__.__name__}.on_connect",
            c="green",
        )

        group_name = websocket.query_params.get(
            "group_name",
        )

        if group_name:
            channel = Channel(
                websocket,
                expires=60 * 60,
                payload_type="json",
            )

            await ChannelBox.add_channel_to_group(
                channel=channel,
                group_name=group_name,
            )

        await websocket.accept()

    async def on_receive(
        self,
        websocket,
        data,
    ):
        sprint(
            f"{self.__class__.__name__}.on_receive {data=} {websocket.query_params=}",
            c="green",
        )

        data = json.loads(
            data,
        )

        message = data["message"]

        username = data["username"]

        if message.strip():
            payload = {
                "username": username,
                "message": message,
            }

            group_name = websocket.query_params.get(
                "group_name",
            )

            if group_name:
                await ChannelBox.group_send(
                    group_name=group_name,
                    payload=payload,
                    save_history=True,
                )


class Chat(HTTPEndpoint):
    async def get(
        self,
        request,
    ):
        sprint(
            f"{self.__class__.__name__}",
            c="green",
        )

        template = Template(
            main_template,
        )

        return HTMLResponse(
            template.render(),
        )


class Chat1(HTTPEndpoint):
    async def get(
        self,
        request,
    ):
        sprint(
            f"{self.__class__.__name__}",
            c="green",
        )

        template = Template(
            chat_template,
        )

        return HTMLResponse(
            template.render(
                SOCKET="127.0.0.1:8888",
                GROUP_NAME="MyChat1",
            ),
        )


class Chat2(HTTPEndpoint):
    async def get(
        self,
        request,
    ):
        sprint(
            f"{self.__class__.__name__}",
            c="green",
        )

        template = Template(
            chat_template,
        )

        return HTMLResponse(
            template.render(
                SOCKET="127.0.0.1:8888",
                GROUP_NAME="MyChat2",
            ),
        )


class SendMessageFromAnyPartOfYourCode(HTTPEndpoint):
    async def get(
        self,
        request,
    ):
        sprint(
            f"{self.__class__.__name__}",
            c="green",
        )

        await ChannelBox.group_send(
            group_name="MyChat",
            payload={
                "username": "Any part of your code",
                "message": "Hello World",
            },
            save_history=True,
        )

        return JSONResponse(
            {
                "message": "success",
            },
        )


class ShowGroups(HTTPEndpoint):
    async def get(
        self,
        request,
    ):
        sprint(
            f"{self.__class__.__name__}",
            c="green",
        )

        groups = await ChannelBox.get_groups()

        return JSONResponse(
            groups,
        )


class FlushGroups(HTTPEndpoint):
    async def get(
        self,
        request,
    ):
        sprint(
            f"{self.__class__.__name__}",
            c="green",
        )

        await ChannelBox.flush_groups()

        return JSONResponse(
            {
                "flush": "success",
            },
        )


class ShowHistory(HTTPEndpoint):
    async def get(
        self,
        request,
    ):
        sprint(
            f"{self.__class__.__name__}",
            c="green",
        )

        history = await ChannelBox.get_history()

        return JSONResponse(
            history,
        )


class FlushHistory(HTTPEndpoint):
    async def get(
        self,
        request,
    ):
        sprint(
            f"{self.__class__.__name__}",
            c="green",
        )

        await ChannelBox.flush_history()

        return JSONResponse(
            {
                "flush": "success",
            },
        )


class CleanExpired(HTTPEndpoint):
    async def get(
        self,
        request,
    ):
        sprint(
            f"{self.__class__.__name__}",
            c="green",
        )

        await ChannelBox.clean_expired()

        return JSONResponse(
            {
                "clean_expired": "success",
            },
        )



main_template = """
<!DOCTYPE html>
<html>
    <head>
        <title>Open this chat in different browsers</title>
        <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-EVSTQN3/azprG1Anm3QDgpJLIm9Nao0Yz1ztcQTwFspd3yD65VohhpuuCOmLASjC" crossorigin="anonymous">
    </head>
    <body>
        <div class="container mt-4">        
            <div class="card">
                <div class="card-header">
                    <ul>
                        <li><a href="/chat1" target="_blank">Chat 1</a></li>
                        <li><a href="/chat2" target="_blank">Chat 2</a></li>
                    </ul>
                </div>
            </div>
        </div>
    </body>
</html>
"""

chat_template = """
<!DOCTYPE html>
<html>
    <head>
        <title>Open this chat in different browsers</title>
        <link href="https://cdn.jsdelivr.net/npm/bootstrap@5.0.2/dist/css/bootstrap.min.css" rel="stylesheet" integrity="sha384-EVSTQN3/azprG1Anm3QDgpJLIm9Nao0Yz1ztcQTwFspd3yD65VohhpuuCOmLASjC" crossorigin="anonymous">
    </head>
    <body>
        <div class="container mt-4">        
            <div class="card">
                <div class="card-header">
                    <h5 class="mt-2">Open this chat in different browsers</h5>
                    <h6>channel-box == 1.1.3</h6>
                    <ul>
                        <li><a href="http://{{ SOCKET }}/send-message-from-any-part-of-your-code" target="_blank">Send message from any part of your code</a></li>
                        <li><a href="http://{{ SOCKET }}/show-groups" target="_blank">Show groups</a></li>
                        <li><a href="http://{{ SOCKET }}/flush-groups" target="_blank">Flush groups</a></li>
                        <li><a href="http://{{ SOCKET }}/show-history" target="_blank">Show history</a></li>
                        <li><a href="http://{{ SOCKET }}/flush-history" target="_blank">Flush history</a></li>
                        <li><a href="http://{{ SOCKET }}/clean-expired" target="_blank">Clean expired</a></li>
                    </ul>
                </div>
                <div class"card-body">   
                    <div class="p-3">                 
                        <form onsubmit="sendMessage(event)">
                            <div class="mb-3">
                                <label for="username" class="form-label">Username</label>
                                <input type="username" class="form-control" id="username" aria-describedby="username" value="root">
                            </div>
                            <div class="mb-3">
                                <label for="message" class="form-label">Message</label>
                                <input type="message" class="form-control" id="message" aria-describedby="username" value="Lorem ipsum">
                            </div>
                            <button type="submit" class="btn btn-primary">Submit</button>
                        </form>
                    </div>
                    <div class="p-3"> 
                        <div id="messages">
                        </div>
                    </div>
                </div>
            </div>
            <script>
                var ws = new WebSocket("ws://{{ SOCKET }}/chat_ws?group_name={{ GROUP_NAME }}"); 
                ws.onopen = function(event) {
                    console.log('Connected to websocket. Channel {{ GROUP_NAME }} is open now.')
                };
                ws.onmessage = function(event) {
                    console.log('Message received %s', event.data)
                    var messages = document.getElementById('messages');
                    var message = document.createElement('div');
                    var data = JSON.parse(event.data);
                    message.innerHTML = `<strong>${data.username} :</strong> ${data.message}`;
                    messages.appendChild(message);
                };
                function sendMessage(event) {
                    var username = document.getElementById("username");
                    var message = document.getElementById("message");
                    var data = {
                        "username": username.value,
                        "message": message.value,
                    };
                    console.log('Message send %s', data)
                    ws.send(JSON.stringify(data));
                    event.preventDefault();
                }
            </script>
        </div>
    </body>
</html>
"""



import setuptools


__author__ = "Sobolev Andrey <email.asobolev@gmail.com>"
__version__ = "1.1.3"


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="channel-box",
    version=__version__,
    install_requires=["starlette>=0.37.2", "websockets>=12.0"],
    author="Sobolev Andrey",
    url="https://github.com/Sobolev5/channel-box",
    author_email="email.asobolev@gmail.com",
    description="ChannelBox it is a package for FastAPI & Starlette frameworks that allows you to make named webscoket channels.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(
        exclude=[
            ".git",
            ".gitignore",
            "*.tests",
            "*.tests.*",
            "tests.*",
            "tests",
            "example",
        ]
    ),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
)



import logging
import sys
import os
import time
import uuid
import datetime
from .utils import (
    PayloadTypeEnum,
    ChannelMessageDC,
)
from starlette.websockets import WebSocket


logger = logging.getLogger(__name__)


class Channel:
    def __init__(
        self,
        websocket: WebSocket,
        expires: int,
        payload_type: str,
    ) -> None:
        """Main websocket channel class.

        Args:
            websocket (WebSocket): Starlette websocket
            expires (int): Channel ttl in seconds
            payload_type (str): encoding of payload (str, bytes, json)

        """
        assert isinstance(websocket, WebSocket)
        assert isinstance(expires, int)
        assert isinstance(payload_type, str) and payload_type in [
            PayloadTypeEnum.JSON.value,
            PayloadTypeEnum.TEXT.value,
            PayloadTypeEnum.BYTES.value,
        ]

        self.websocket = websocket
        self.expires = expires
        self.payload_type = payload_type
        self.uuid = uuid.uuid4()
        self.last_active = time.time()

    async def _send(self, payload: str) -> None:
        match self.payload_type:
            case PayloadTypeEnum.JSON.value:
                await self.websocket.send_json(payload)

            case PayloadTypeEnum.TEXT.value:
                await self.websocket.send_text(payload)

            case PayloadTypeEnum.BYTES.value:
                await self.websocket.send_bytes(payload)

            case _:
                await self.websocket.send(payload)

        self.last_active = time.time()  # renew last_active time for active connecitons

    async def _is_expired(self) -> bool:
        return (self.expires + int(self.last_active)) < time.time()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {self.uuid} {self.payload_type=} {self.expires=}"


class ChannelBox:
    CHANNEL_GROUPS: dict = {}
    CHANNEL_GROUPS_HISTORY: dict = {}
    HISTORY_SIZE: int = int(os.getenv("CHANNEL_BOX_HISTORY_SIZE", 1_048_576))

    @classmethod
    async def add_channel_to_group(
        cls,
        channel: Channel,
        group_name: str = "default",
    ) -> None:
        """Add channel to group.

        Args:
            channel (Channel): Instance of Channel class
            group_name (str): Group name

        """
        if group_name not in cls.CHANNEL_GROUPS:
            cls.CHANNEL_GROUPS[group_name] = {}

        if channel not in cls.CHANNEL_GROUPS[group_name]:
            cls.CHANNEL_GROUPS[group_name][channel] = {
                "created_at": datetime.datetime.now(tz=datetime.UTC)
            }

    @classmethod
    async def remove_channel_from_group(
        cls,
        channel: Channel,
        group_name: str,
    ) -> None:
        """Remove channel from group.

        Args:
            channel (Channel): Instance of Channel class
            group_name (str): Group name

        """
        if channel in cls.CHANNEL_GROUPS.get(group_name, {}):
            try:
                del cls.CHANNEL_GROUPS[group_name][channel]
            except KeyError:
                logger.warning(
                    "Channel %s not found in group %s during cleanup",
                    channel,
                    group_name,
                )

    @classmethod
    async def group_send(
        cls,
        group_name: str = "default",
        payload: dict | str | bytes = {},
        save_history: bool = False,
    ) -> None:
        """Send payload to all channels connected to group.

        Args:
            group_name (str, optional): Group name
            payload (dict, optional): Payload to channel
            save_history (bool, optional): Save message history. Defaults to False.

        """
        if save_history:
            if group_name not in cls.CHANNEL_GROUPS_HISTORY:
                cls.CHANNEL_GROUPS_HISTORY[group_name] = []

            cls.CHANNEL_GROUPS_HISTORY[group_name].append(
                ChannelMessageDC(
                    payload=payload,
                )
            )

            if sys.getsizeof(cls.CHANNEL_GROUPS_HISTORY[group_name]) > cls.HISTORY_SIZE:
                cls.CHANNEL_GROUPS_HISTORY[group_name] = []

        for channel in cls.CHANNEL_GROUPS.get(group_name, {}):
            await channel._send(payload)

    @classmethod
    async def get_groups(
        cls,
    ) -> dict:
        return cls.CHANNEL_GROUPS

    @classmethod
    async def flush_groups(
        cls,
    ) -> None:
        cls.CHANNEL_GROUPS = {}

    @classmethod
    async def get_history(
        cls,
        group_name: str = "",
    ) -> dict:
        return (
            cls.CHANNEL_GROUPS_HISTORY.get(group_name, {})
            if group_name
            else cls.CHANNEL_GROUPS_HISTORY
        )

    @classmethod
    async def flush_history(
        cls,
    ) -> None:
        cls.CHANNEL_GROUPS_HISTORY = {}

    @classmethod
    async def clean_expired(
        cls,
    ) -> None:
        for group_name in cls.CHANNEL_GROUPS:
            for channel in cls.CHANNEL_GROUPS.get(group_name, {}):
                is_expired = await channel._is_expired()
                if is_expired:
                    try:
                        del cls.CHANNEL_GROUPS[group_name][channel]
                    except KeyError:
                        logger.warning(
                            "Channel %s not found in group %s during cleanup",
                            channel,
                            group_name,
                        )

            if not any(cls.CHANNEL_GROUPS.get(group_name, {})):
                try:
                    del cls.CHANNEL_GROUPS[group_name]
                except KeyError:
                    logger.warning("Group %s not found during cleanup", group_name)
